module(..., package.seeall)

local utils = require "selenium.utils"
local socket = require "socket"
local http = require "socket.http"
local ltn12 = require "ltn12"

module(..., package.seeall)

local methods = {}
local mt = { __index = methods }

function new (host, port, browserStartCommand, browserURL)
	local n = setmetatable({}, mt)
	
	n.host = host
	n.port = port
	n.browserStartCommand = browserStartCommand
	n.browserURL = browserURL
	n.sessionId = nil
	n.extensionJs = ""
	n.verbose = false
	
	return n
end

---
--
function methods:doCommand (verb, args)
	local params = {
		cmd = verb
	}
	
	if type(args) == "table" then
		for index, arg in ipairs(args) do
			params[index] = arg
		end
	end
	
	if self.verbose then
		print( ("[selenium] %s %s"):format(verb, table.concat(args or {}, " | ")) )
	end
	
	if self.sessionId ~= nil then
		params.sessionId = self.sessionId
	end
	
	local urlParams = utils.encode(params)
	local heads = {
		["Content-Type"] = "application/x-www-form-urlencoded; charset=utf-8", 
		["Content-Length"] = tostring(#urlParams)
	}
	
	local tbl = {}
	local r, code = http.request{ 
	    url = ("http://%s:%d/selenium-server/driver/"):format(self.host, self.port), 
		sink = ltn12.sink.table(tbl),
		method = "POST",
		headers = heads,
		source = ltn12.source.string(urlParams)
	}

	if code ~= 200 then
		print(code)
		error("Response from selenium server is Wrong")
	end
	if self.sessionId == nil then
		self.sessionId = tbl[1]:sub(4)
	end
	return tbl[1]
end

---
-- 
function methods:setVerbose ()
	self.verbose = true
end

---
--
function methods:unsetVerbose ()
	self.verbose = false
end

---
--	
function methods:setExtensionJs (extensionJs)
	self.extensionJs = extensionJs
end


function methods:start ()
	local result = self:getString("getNewBrowserSession", {self.browserStartCommand, self.browserURL, self.extensionJs})
	self.sessionId = result
end

function methods:stop ()
	self:doCommand("testComplete", {})
	self.sessionId = nil
end

---
--
function methods:waitForElement (options)
	-- locator is required, timeout optional 
	if options.locator == nil then
		error("No locator provided for wait_for_element")
	end	
	local timeout = options.timeout or 3
	while timeout > 0 do
		timeout = timeout - 0.250
		socket.sleep(0.250)
		
		if self:isElementPresent(options.locator) then
			return true
		end	
	end		
	return false
end

function methods:waitForElementNotPresent (options)
	if options.locator == nil then
		error("Not location provided for wait_for_element")
	end	
	
	local timeout = options.timeout or 3
	while timeout > 0 do
		timeout = timeout - 0.250
		socket.sleep(0.250)
		if not self:isElementPresent(options.locator) then
			return true
		end
	end	
	return false	
end

function methods:getString (verb, args)
	local result = self:doCommand(verb, args)
	return result:sub(4)
end

function methods:getStringArray (verb, args)
	local csv = self:getString(verb, args)
    local token = ""
    local tokens = {}
    local escape = false
	
	for i = 1, #csv do
		local continue = false
		local letter = csv:sub(i, i)
		if escape then
			token  = token .. letter
			escape = false
			continue = true
		end
		if not continue then
			if letter == '\\' then
				escape = true
			elseif letter == ',' then
				tokens[#tokens + 1] = token
				token = ''
			else
				token = token .. letter
			end
		end
	end
	tokens[#tokens + 1] = token
	return tokens
end	

function methods:getNumber (verb, args)
	return self:getString(verb, args)
end

function methods:getNumberArray (verb, args)
	return self:getStringArray(verb, args)
end

function methods:getBoolean (verb, args)
    local boolstr = self:getString(verb, args)
    if boolstr == "true" then
		return true
    elseif boolstr == "false" then
		return false
	end
	error("result is neither 'true' nor 'false': " .. tostring(boolstr))
end

function methods:getBooleanArray (verb, args)
    local boolarr = self:getStringArray(verb, args)
    for i = 1, #boolarr do
		if boolstr == "true" then
			boolarr[i] = true
		elseif boolstr == "false" then
    		boolarr[i] = false
		else
			error("result is neither 'true' nor 'false': " .. tostring(boolarr[i]))
		end
    end
	return boolarr
end

--- From here on, everything's auto-generated from XML

--[[
	Clicks on a link, button, checkbox or radio button. If the click action
	causes a new page to load (like a link usually does), call
	waitForPageToLoad.
  
	'locator' is an element locator
	--]]
function methods:click (locator)
	self:doCommand("click", {locator})
end

--[[
Double clicks on a link, button, checkbox or radio button. If the double click action
causes a new page to load (like a link usually does), call
waitForPageToLoad.
 
'locator' is an element locator
--]]
function methods:doubleClick (locator)
	self:doCommand("doubleClick", {locator})
end

--[[
Simulates opening the context menu for the specified element (as might happen if the user "right-clicked" on the element).

'locator' is an element locator
--]]
function methods:contextMenu (locator)
	self:doCommand("contextMenu", {locator})
end

--[[
Clicks on a link, button, checkbox or radio button. If the click action
causes a new page to load (like a link usually does), call
waitForPageToLoad.

'locator' is an element locator
'coordString' is specifies the x,y position (i.e. - 10,20) of the mouse      event relative to the element returned by the locator.
--]]
function methods:clickAt (locator, coordString)
	self:doCommand("clickAt", {locator, coordString})
end

--[[
Doubleclicks on a link, button, checkbox or radio button. If the action
causes a new page to load (like a link usually does), call
waitForPageToLoad.

'locator' is an element locator
'coordString' is specifies the x,y position (i.e. - 10,20) of the mouse      event relative to the element returned by the locator.
--]]
function methods:doubleClickAt (locator, coordString)
	self:doCommand("doubleClickAt", {locator, coordString})
end

--[[
Simulates opening the context menu for the specified element (as might happen if the user "right-clicked" on the element).

'locator' is an element locator
'coordString' is specifies the x,y position (i.e. - 10,20) of the mouse      event relative to the element returned by the locator.
--]]
function methods:contextMenuAt (locator, coordString)
	self:doCommand("contextMenuAt", {locator, coordString})
end

--[[
Explicitly simulate an event, to trigger the corresponding "on\ *event*"
handler.

'locator' is an element locator
'eventName' is the event name, e.g. "focus" or "blur"
--]]
function methods:fireEvent (locator, eventName)
	self:doCommand("fireEvent", {locator, eventName})
end

--[[
Move the focus to the specified element; for example, if the element is an input field, move the cursor to that field.

'locator' is an element locator
--]]
function methods:focus (locator)
	self:doCommand("focus", {locator})
end

--[[
Simulates a user pressing and releasing a key.

'locator' is an element locator
'keySequence' is Either be a string("\" followed by the numeric keycode  of the key to be pressed, normally the ASCII value of that key), or a single  character. For example: "w", "\119".
--]]
function methods:keyPress (locator, keySequence)
	self:doCommand("keyPress", {locator, keySequence})
end

--[[
Press the shift key and hold it down until doShiftUp() is called or a new page is loaded.

--]]
function methods:shiftKeyDown ()
	self:doCommand("shiftKeyDown", {})
end

--[[
Release the shift key.

--]]
function methods:shiftKeyUp ()
	self:doCommand("shiftKeyUp", {})
end

--[[
Press the meta key and hold it down until doMetaUp() is called or a new page is loaded.

--]]
function methods:metaKeyDown ()
	self:doCommand("metaKeyDown", {})
end

--[[
Release the meta key.

--]]
function methods:metaKeyUp ()
	self:doCommand("metaKeyUp", {})
end

--[[
Press the alt key and hold it down until doAltUp() is called or a new page is loaded.

--]]
function methods:altKeyDown ()
	self:doCommand("altKeyDown", {})
end

--[[
Release the alt key.

--]]
function methods:altKeyUp ()
	self:doCommand("altKeyUp", {})
end

--[[
Press the control key and hold it down until doControlUp() is called or a new page is loaded.

--]]
function methods:controlKeyDown ()
	self:doCommand("controlKeyDown", {})
end

--[[
Release the control key.

--]]
function methods:controlKeyUp ()
	self:doCommand("controlKeyUp", {})
end

--[[
Simulates a user pressing a key (without releasing it yet).

'locator' is an element locator
'keySequence' is Either be a string("\" followed by the numeric keycode  of the key to be pressed, normally the ASCII value of that key), or a single  character. For example: "w", "\119".
--]]
function methods:keyDown (locator, keySequence)
	self:doCommand("keyDown", {locator, keySequence})
end

--[[
Simulates a user releasing a key.
'locator' is an element locator
'keySequence' is Either be a string("\" followed by the numeric keycode  of the key to be pressed, normally the ASCII value of that key), or a single  character. For example: "w", "\119".
--]]
function methods:keyUp (locator, keySequence)
	self:doCommand("keyUp", {locator, keySequence})
end

--[[
Simulates a user hovering a mouse over the specified element.

'locator' is an element locator
--]]
function methods:mouseOver (locator)
	self:doCommand("mouseOver", {locator})
end

--[[
Simulates a user moving the mouse pointer away from the specified element.

'locator' is an element locator
--]]
function methods:mouseOut (locator)
	self:doCommand("mouseOut", {locator})
end

--[[
Simulates a user pressing the left mouse button (without releasing it yet) on
the specified element.

'locator' is an element locator
--]]
function methods:mouseDown (locator)
	self:doCommand("mouseDown", {locator})
end

--[[
Simulates a user pressing the right mouse button (without releasing it yet) on
the specified element.

'locator' is an element locator
--]]
function methods:mouseDownRight (locator)
	self:doCommand("mouseDownRight", {locator})
end

--[[
Simulates a user pressing the left mouse button (without releasing it yet) at
the specified location.

'locator' is an element locator
'coordString' is specifies the x,y position (i.e. - 10,20) of the mouse      event relative to the element returned by the locator.
--]]
function methods:mouseDownAt (locator, coordString)
	self:doCommand("mouseDownAt", {locator, coordString})
end

--[[
Simulates a user pressing the right mouse button (without releasing it yet) at
the specified location.

'locator' is an element locator
'coordString' is specifies the x,y position (i.e. - 10,20) of the mouse      event relative to the element returned by the locator.
--]]
function methods:mouseDownRightAt (locator, coordString)
	self:doCommand("mouseDownRightAt", {locator, coordString})
end

--[[
Simulates the event that occurs when the user releases the mouse button (i.e., stops
holding the button down) on the specified element.

'locator' is an element locator
--]]
function methods:mouseUp (locator)
	self:doCommand("mouseUp", {locator})
end

--[[
Simulates the event that occurs when the user releases the right mouse button (i.e., stops
holding the button down) on the specified element.

'locator' is an element locator
--]]
function methods:mouseUpRight (locator)
	self:doCommand("mouseUpRight", {locator})
end

--[[
Simulates the event that occurs when the user releases the mouse button (i.e., stops
holding the button down) at the specified location.

'locator' is an element locator
'coordString' is specifies the x,y position (i.e. - 10,20) of the mouse      event relative to the element returned by the locator.
--]]
function methods:mouseUpAt (locator, coordString)
	self:doCommand("mouseUpAt", {locator, coordString})
end

--[[
Simulates the event that occurs when the user releases the right mouse button (i.e., stops
holding the button down) at the specified location.

'locator' is an element locator
'coordString' is specifies the x,y position (i.e. - 10,20) of the mouse      event relative to the element returned by the locator.
--]]
function methods:mouseUpRightAt (locator, coordString)
	self:doCommand("mouseUpRightAt", {locator, coordString})
end

--[[
Simulates a user pressing the mouse button (without releasing it yet) on
the specified element.

'locator' is an element locator
--]]
function methods:mouseMove (locator)
	self:doCommand("mouseMove", {locator})
end

--[[
Simulates a user pressing the mouse button (without releasing it yet) on
the specified element.

'locator' is an element locator
'coordString' is specifies the x,y position (i.e. - 10,20) of the mouse      event relative to the element returned by the locator.
--]]
function methods:mouseMoveAt (locator, coordString)
	self:doCommand("mouseMoveAt", {locator, coordString})
end

--[[
Sets the value of an input field, as though you typed it in.


Can also be used to set the value of combo boxes, check boxes, etc. In these cases,
value should be the value of the option selected, not the visible text.


'locator' is an element locator
'value' is the value to type
--]]
function methods:type (locator, value)
	self:doCommand("type", {locator, value})
end

--[[
Simulates keystroke events on the specified element, as though you typed the value key-by-key.


This is a convenience method for calling keyDown, keyUp, keyPress for every character in the specified string;
this is useful for dynamic UI widgets (like auto-completing combo boxes) that require explicit key events.

Unlike the simple "type" command, which forces the specified value into the page directly, this command
may or may not have any visible effect, even in cases where typing keys would normally have a visible effect.
For example, if you use "typeKeys" on a form element, you may or may not see the results of what you typed in
the field.

In some cases, you may need to use the simple "type" command to set the value of the field and then the "typeKeys" command to
send the keystroke events corresponding to what you just typed.


'locator' is an element locator
'value' is the value to type
--]]
function methods:typeKeys (locator, value)
	self:doCommand("typeKeys", {locator, value})
end

--[[
Set execution speed (i.e., set the millisecond length of a delay which will follow each selenium operation).  By default, there is no such delay, i.e.,
the delay is 0 milliseconds.

'value' is the number of milliseconds to pause after operation
--]]
function methods:setSpeed (value)
	self:doCommand("setSpeed", {value})
end

--[[
Get execution speed (i.e., get the millisecond length of the delay following each selenium operation).  By default, there is no such delay, i.e.,
the delay is 0 milliseconds.

See also setSpeed.

--]]
function methods:getSpeed ()
	return self:getString("getSpeed", {})
end

--[[
Check a toggle-button (checkbox/radio)

'locator' is an element locator
--]]
function methods:check (locator)
	self:doCommand("check", {locator})
end

--[[
Uncheck a toggle-button (checkbox/radio)

'locator' is an element locator
--]]
function methods:uncheck (locator)
	self:doCommand("uncheck", {locator})
end

--[[
Select an option from a drop-down using an option locator.
--]]
function methods:select (selectLocator, optionLocator)
	self:doCommand("select", {selectLocator, optionLocator})
end

--[[
Add a selection to the set of selected options in a multi-select element using an option locator.

@see #doSelect for details of option locators

'locator' is an element locator identifying a multi-select box
'optionLocator' is an option locator (a label by default)
--]]
function methods:addSelection (locator, optionLocator)
	self:doCommand("addSelection", {locator, optionLocator})
end

--[[
Remove a selection from the set of selected options in a multi-select element using an option locator.

@see #doSelect for details of option locators

'locator' is an element locator identifying a multi-select box
'optionLocator' is an option locator (a label by default)
--]]
function methods:removeSelection (locator, optionLocator)
	self:doCommand("removeSelection", {locator, optionLocator})
end

--[[
Unselects all of the selected options in a multi-select element.

'locator' is an element locator identifying a multi-select box
--]]
function methods:removeAllSelections (locator)
	self:doCommand("removeAllSelections", {locator})
end

--[[
Submit the specified form. This is particularly useful for forms without
submit buttons, e.g. single-input "Search" forms.

'formLocator' is an element locator for the form you want to submit
--]]
function methods:submit (formLocator)
	self:doCommand("submit", {formLocator})
end	

--[[
Opens an URL in the test frame. This accepts both relative and absolute
URLs.

The 'open' command waits for the page to load before proceeding,
ie. the 'AndWait' suffix is implicit.

\ *Note*: The URL must be on the same domain as the runner HTML
due to security restrictions in the browser (Same Origin Policy). If you
need to open an URL on another domain, use the Selenium Server to start a
new browser session on that domain.

'url' is the URL to open; may be relative or absolute
--]]
function methods:open (url)
	self:doCommand("open", {url})
end

--[[
Opens a popup window (if a window with that ID isn't already open).
After opening the window, you'll need to select it using the selectWindow
command.


This command can also be a useful workaround for bug SEL-339.  In some cases, Selenium will be unable to intercept a call to window.open (if the call occurs during or before the "onLoad" event, for example).
In those cases, you can force Selenium to notice the open window's name by using the Selenium openWindow command, using
an empty (blank) url, like this: openWindow("", "myFunnyWindow").


'url' is the URL to open, which can be blank
'windowID' is the JavaScript window ID of the window to select
--]]
function methods:openWindow (url, windowID)
	self:doCommand("openWindow", {url, windowID})
end

function methods:selectWindow (windowID)
	self:doCommand("selectWindow", {windowID})
end

--[[
Simplifies the process of selecting a popup window (and does not offer
functionality beyond what ``selectWindow()`` already provides).

	*   If ``windowID`` is either not specified, or specified as
	"null", the first non-top window is selected. The top window is the one
	that would be selected by ``selectWindow()`` without providing a
	``windowID`` . This should not be used when more than one popup
	window is in play.
	*   Otherwise, the window will be looked up considering
	``windowID`` as the following in order: 1) the "name" of the
	window, as specified to ``window.open()``; 2) a javascript
	variable which is a reference to a window; and 3) the title of the
	window. This is the same ordered lookup performed by
	``selectWindow`` .

'windowID' is an identifier for the popup window, which can take on a  number of different meanings
--]]
function methods:selectPopUp (windowID)
	self:doCommand("selectPopUp", {windowID})
end

--[[
Selects the main window. Functionally equivalent to using
``selectWindow()`` and specifying no value for
``windowID``.
--]]
function methods:deselectPopUp ()
	self:doCommand("deselectPopUp", {})
end

--[[
Selects a frame within the current window.  (You may invoke this command
multiple times to select nested frames.)  To select the parent frame, use
'relative=parent' as a locator; to select the top frame, use 'relative=top'.
You can also select a frame by its 0-based index number; select the first frame with
'index=0', or the third frame with 'index=2'.


You may also use a DOM expression to identify the frame you want directly,
like this: ``dom=frames['main'].frames['subframe']``

'locator' is an element locator identifying a frame or iframe
--]]
function methods:selectFrame (locator)
	self:doCommand("selectFrame", {locator})
end

--[[
Determine whether current/locator identify the frame containing this running code.

This is useful in proxy injection mode, where this code runs in every
browser frame and window, and sometimes the selenium server needs to identify
the 'current' frame.  In this case, when the test calls selectFrame, this
routine is called for each frame to figure out which one has been selected.
The selected frame will return true, while all others will return false.


'currentFrameString' is starting frame
'target' is new frame (which might be relative to the current one)
--]]
function methods:getWhetherThisFrameMatchFrameExpression (currentFrameString, target)
	return self:getBoolean("getWhetherThisFrameMatchFrameExpression", {currentFrameString, target})
end

--[[
Determine whether currentWindowString plus target identify the window containing this running code.


This is useful in proxy injection mode, where this code runs in every
browser frame and window, and sometimes the selenium server needs to identify
the "current" window.  In this case, when the test calls selectWindow, this
routine is called for each window to figure out which one has been selected.
The selected window will return true, while all others will return false.


'currentWindowString' is starting window
'target' is new window (which might be relative to the current one, e.g., "_parent")
--]]
function methods:getWhetherThisWindowMatchWindowExpression (currentWindowString, target)
	return self:getBoolean("getWhetherThisWindowMatchWindowExpression", {currentWindowString, target})
end

--[[
Waits for a popup window to appear and load up.

'windowID' is the JavaScript window "name" of the window that will appear (not the text of the title bar) If unspecified, or specified as "null", this command will wait for the first non-top window to appear (don't rely on this if you are working with multiple popups simultaneously).
'timeout' is a timeout in milliseconds, after which the action will return with an error.If this value is not specified, the default Seleniumtimeout will be used. See the setTimeout() command.
--]]
function methods:waitForPopUp (windowID, timeout)
	self:doCommand("waitForPopUp", {windowID, timeout})
end

function methods:chooseCancelOnNextConfirmation ()
	self:doCommand("chooseCancelOnNextConfirmation", {})
end

--[[
Undo the effect of calling chooseCancelOnNextConfirmation.  Note
that Selenium's overridden window.confirm() function will normally automatically
return true, as if the user had manually clicked OK, so you shouldn't
need to use this command unless for some reason you need to change
your mind prior to the next confirmation.  After any confirmation, Selenium will resume using the
default behavior for future confirmations, automatically returning 
true (OK) unless/until you explicitly call chooseCancelOnNextConfirmation for each
confirmation.

Take note - every time a confirmation comes up, you must
consume it with a corresponding getConfirmation, or else
the next selenium operation will fail.
--]]
function methods:chooseOkOnNextConfirmation ()
	self:doCommand("chooseOkOnNextConfirmation", {})
end

--[[
Instructs Selenium to return the specified answer string in response to
the next JavaScript prompt [window.prompt()].

'answer' is the answer to give in response to the prompt pop-up
--]]
function methods:answerOnNextPrompt (answer)
	self:doCommand("answerOnNextPrompt", {answer})
end

--[[
Simulates the user clicking the "back" button on their browser.
--]]
function methods:goBack ()
	self:doCommand("goBack", {})
end

--[[
Simulates the user clicking the 'Refresh' button on their browser.
--]]
function methods:refresh ()
	self:doCommand("refresh", {})
end

--[[
Simulates the user clicking the "close" button in the titlebar of a popup
window or tab.
--]]
function methods:close ()
	self:doCommand("close", {})
end

--[[
Has an alert occurred?
This function never throws an exception
--]]
function methods:isAlertPresent ()
	return self:getBoolean("isAlertPresent", {})
end

--[[
Has a prompt occurred?
This function never throws an exception
--]]
function methods:isPromptPresent ()
	return self:getBoolean("isPromptPresent", {})
end

--[[
Has confirm() been called?
This function never throws an exception
--]]
function methods:isConfirmationPresent ()
	return self:getBoolean("isConfirmationPresent", {})
end


--[[
Retrieves the message of a JavaScript alert generated during the previous action, or fail if there were no alerts.


Getting an alert has the same effect as manually clicking OK. If an
alert is generated but you do not consume it with getAlert, the next Selenium action
will fail.

Under Selenium, JavaScript alerts will NOT pop up a visible alert
dialog.

Selenium does NOT support JavaScript alerts that are generated in a
page's onload() event handler. In this case a visible dialog WILL be
generated and Selenium will hang until someone manually clicks OK.
--]]
function methods:getAlert ()
	return self:getString("getAlert", {})
end

--[[
Retrieves the message of a JavaScript confirmation dialog generated during
the previous action.

By default, the confirm function will return true, having the same effect
as manually clicking OK. This can be changed by prior execution of the
chooseCancelOnNextConfirmation command. 

If an confirmation is generated but you do not consume it with getConfirmation,
the next Selenium action will fail.

NOTE: under Selenium, JavaScript confirmations will NOT pop up a visible
dialog.

NOTE: Selenium does NOT support JavaScript confirmations that are
generated in a page's onload() event handler. In this case a visible
dialog WILL be generated and Selenium will hang until you manually click
OK.
--]]
function methods:getConfirmation ()
	return self:getString("getConfirmation", {})
end

--[[
Retrieves the message of a JavaScript question prompt dialog generated during
the previous action.

Successful handling of the prompt requires prior execution of the
answerOnNextPrompt command. If a prompt is generated but you
do not get/verify it, the next Selenium action will fail.

NOTE: under Selenium, JavaScript prompts will NOT pop up a visible
dialog.

NOTE: Selenium does NOT support JavaScript prompts that are generated in a
page's onload() event handler. In this case a visible dialog WILL be
generated and Selenium will hang until someone manually clicks OK.
--]]
function methods:getPrompt ()
	return self:getString("getPrompt", {})
end

--[[
Gets the absolute URL of the current page.
--]]
function methods:getLocation ()
	return self:getString("getLocation", {})
end

--[[
Gets the title of the current page.
--]]
function methods:getTitle ()
	return self:getString("getTitle", {})
end

--[[
Gets the entire text of the page.
--]]
function methods:getBodyText ()
	return self:getString("getBodyText", {})
end

function methods:getValue (locator)
	return self:getString("getValue", {locator})
end

--[[
Gets the text of an element. This works for any element that contains
text. This command uses either the textContent (Mozilla-like browsers) or
the innerText (IE-like browsers) of the element, which is the rendered
text shown to the user.

'locator' is an element locator
--]]
function methods:getText (locator)
	return self:getString("getText", {locator})
end

function methods:highlight (locator)
	self:doCommand("highlight", {locator})
end

--[[
Gets the result of evaluating the specified JavaScript snippet.  The snippet may
have multiple lines, but only the result of the last line will be returned.


Note that, by default, the snippet will run in the context of the "selenium"
object itself, so ``this`` will refer to the Selenium object.  Use ``window`` to
refer to the window of your application, e.g. ``window.document.getElementById('foo')``

If you need to use
a locator to refer to a single element in your application page, you can
use ``this.browserbot.findElement("id=foo")`` where "id=foo" is your locator.


'script' is the JavaScript snippet to run
--]]
function methods:getEval (script)
	return self:getString("getEval", {script})
end

--[[
Gets whether a toggle-button (checkbox/radio) is checked.  Fails if the specified element doesn't exist or isn't a toggle-button.

'locator' is an element locator pointing to a checkbox or radio button
--]]
function methods:isChecked (locator)
	return self:getBoolean("isChecked", {locator})
end

--[[
Gets the text from a cell of a table. The cellAddress syntax
tableLocator.row.column, where row and column start at 0.

'tableCellAddress' is a cell address, e.g. "foo.1.4"
--]]
function methods:getTable (tableCellAddress)
	return self:getString("getTable", {tableCellAddress})
end

--[[
Gets all option labels (visible text) for selected options in the specified select or multi-select element.

'selectLocator' is an element locator identifying a drop-down menu
--]]
function methods:getSelectedLabels (selectLocator)
	return self:getStringArray("getSelectedLabels", {selectLocator})
end

--[[
Gets option label (visible text) for selected option in the specified select element.

'selectLocator' is an element locator identifying a drop-down menu
--]]
function methods:getSelectedLabel (selectLocator)
	return self:getString("getSelectedLabel", {selectLocator})
end

--[[
Gets all option values (value attributes) for selected options in the specified select or multi-select element.

'selectLocator' is an element locator identifying a drop-down menu
--]]
function methods:getSelectedValues (selectLocator)
	return self:getStringArray("getSelectedValues", {selectLocator})
end

--[[
Gets option value (value attribute) for selected option in the specified select element.

'selectLocator' is an element locator identifying a drop-down menu
--]]
function methods:getSelectedValue (selectLocator)
	return self:getString("getSelectedValue", {selectLocator})
end

--[[
Gets all option indexes (option number, starting at 0) for selected options in the specified select or multi-select element.

'selectLocator' is an element locator identifying a drop-down menu
--]]
function methods:getSelectedIndexes (selectLocator)
	return self:getStringArray("getSelectedIndexes", {selectLocator})
end

--[[
Gets option index (option number, starting at 0) for selected option in the specified select element.

'selectLocator' is an element locator identifying a drop-down menu
--]]
function methods:getSelectedIndex (selectLocator)
	return self:getString("getSelectedIndex", {selectLocator})
end

--[[
Gets all option element IDs for selected options in the specified select or multi-select element.

'selectLocator' is an element locator identifying a drop-down menu
--]]
function methods:getSelectedIds (selectLocator)
	return self:getStringArray("getSelectedIds", {selectLocator})
end

--[[
Gets option element ID for selected option in the specified select element.

'selectLocator' is an element locator identifying a drop-down menu
--]]
function methods:getSelectedId (selectLocator)
	return self:getString("getSelectedId", {selectLocator})
end

--[[
Determines whether some option in a drop-down menu is selected.

'selectLocator' is an element locator identifying a drop-down menu
--]]
function methods:isSomethingSelected (selectLocator)
	return self:getBoolean("isSomethingSelected", {selectLocator})
end

--[[
Gets all option labels in the specified select drop-down.

'selectLocator' is an element locator identifying a drop-down menu
--]]
function methods:getSelectOptions (selectLocator)
	return self:getStringArray("getSelectOptions", {selectLocator})
end

--[[
Gets the value of an element attribute. The value of the attribute may
differ across browsers (this is the case for the "style" attribute, for
example).

'attributeLocator' is an element locator followed by an @ sign and then the name of the attribute, e.g. "foo@bar"
--]]
function methods:getAttribute (attributeLocator)
	return self:getString("getAttribute", {attributeLocator})
end

--[[
Verifies that the specified text pattern appears somewhere on the rendered page shown to the user.

'pattern' is a pattern to match with the text of the page
--]]
function methods:isTextPresent (pattern)
	return self:getBoolean("isTextPresent", {pattern})
end

--[[
Verifies that the specified element is somewhere on the page.

'locator' is an element locator
--]]
function methods:isElementPresent (locator)
	return self:getBoolean("isElementPresent", {locator})
end

--[[
Determines if the specified element is visible. An
element can be rendered invisible by setting the CSS "visibility"
property to "hidden", or the "display" property to "none", either for the
element itself or one if its ancestors.  This method will fail if
the element is not present.

'locator' is an element locator
--]]
function methods:isVisible (locator)
	return self:getBoolean("isVisible", {locator})
end

--[[
Determines whether the specified input element is editable, ie hasn't been disabled.
This method will fail if the specified element isn't an input element.

'locator' is an element locator
--]]
function methods:isEditable (locator)
	return self:getBoolean("isEditable", {locator})
end

--[[
Returns the IDs of all buttons on the page.
If a given button has no ID, it will appear as "" in this array.
--]]
function methods:getAllButtons ()
	return self:getStringArray("getAllButtons", {})
end

--[[
Returns the IDs of all links on the page.
If a given link has no ID, it will appear as "" in this array.
--]]
function methods:getAllLinks ()
	return self:getStringArray("getAllLinks", {})
end

--[[
Returns the IDs of all input fields on the page.
If a given field has no ID, it will appear as "" in this array.
--]]
function methods:getAllFields ()
	return self:getStringArray("getAllFields", {})
end

--[[
Returns every instance of some attribute from all known windows.

'attributeName' is name of an attribute on the windows
--]]
function methods:getAttributeFromAllWindows (attributeName)
	return self:getStringArray("getAttributeFromAllWindows", {attributeName})
end

--[[
deprecated - use dragAndDrop instead

'locator' is an element locator
'movementsString' is offset in pixels from the current location to which the element should be moved, e.g., "+70,-300"
--]]
function methods:dragdrop (locator, movementsString)
	self:doCommand("dragdrop", {locator, movementsString})
end

--[[
Configure the number of pixels between "mousemove" events during dragAndDrop commands (default=10).

Setting this value to 0 means that we'll send a "mousemove" event to every single pixel
in between the start location and the end location; that can be very slow, and may
cause some browsers to force the JavaScript to timeout.

If the mouse speed is greater than the distance between the two dragged objects, we'll
just send one "mousemove" at the start location and then one final one at the end location.


'pixels' is the number of pixels between "mousemove" events
--]]
function methods:setMouseSpeed (pixels)
	self:doCommand("setMouseSpeed", {pixels})
end

--[[
Returns the number of pixels between "mousemove" events during dragAndDrop commands (default=10).
--]]
function methods:getMouseSpeed ()
	return self:getNumber("getMouseSpeed", {})
end

--[[
Drags an element a certain distance and then drops it

'locator' is an element locator
'movementsString' is offset in pixels from the current location to which the element should be moved, e.g., "+70,-300"
--]]
function methods:dragAndDrop (locator, movementsString)
	self:doCommand("dragAndDrop", {locator, movementsString})
end

--[[
Drags an element and drops it on another element

'locatorOfObjectToBeDragged' is an element to be dragged
'locatorOfDragDestinationObject' is an element whose location (i.e., whose center-most pixel) will be the point where locatorOfObjectToBeDragged  is dropped
--]]
function methods:dragAndDropToObject (locatorOfObjectToBeDragged, locatorOfDragDestinationObject)
	self:doCommand("dragAndDropToObject", {locatorOfObjectToBeDragged, locatorOfDragDestinationObject})
end

--[[
Gives focus to the currently selected window
--]]
function methods:windowFocus ()
	self:doCommand("windowFocus", {})
end

--[[
Resize currently selected window to take up the entire screen
--]]
function methods:windowMaximize ()
	self:doCommand("windowMaximize", {})
end

--[[
Returns the IDs of all windows that the browser knows about.
--]]
function methods:getAllWindowIds ()
	return self:getStringArray("getAllWindowIds", {})
end

--[[
Returns the names of all windows that the browser knows about.
--]]
function methods:getAllWindowNames ()
	return self:getStringArray("getAllWindowNames", {})
end

--[[
Returns the titles of all windows that the browser knows about.
--]]
function methods:getAllWindowTitles ()
	return self:getStringArray("getAllWindowTitles", {})
end

--[[
Returns the entire HTML source between the opening and
closing "html" tags.
--]]
function methods:getHtmlSource ()
	return self:getString("getHtmlSource", {})
end

--[[
Moves the text cursor to the specified position in the given input element or textarea.
This method will fail if the specified element isn't an input element or textarea.

'locator' is an element locator pointing to an input element or textarea
'position' is the numerical position of the cursor in the field; position should be 0 to move the position to the beginning of the field.  You can also set the cursor to -1 to move it to the end of the field.
--]]
function methods:setCursorPosition (locator, position)
	self:doCommand("setCursorPosition", {locator, position})
end

--[[
Get the relative index of an element to its parent (starting from 0). The comment node and empty text node
will be ignored.

'locator' is an element locator pointing to an element
--]]
function methods:getElementIndex (locator)
	return self:getNumber("getElementIndex", {locator})
end

--[[
Check if these two elements have same parent and are ordered siblings in the DOM. Two same elements will
not be considered ordered.

'locator1' is an element locator pointing to the first element
'locator2' is an element locator pointing to the second element
--]]
function methods:isOrdered (locator1, locator2)
	return self:getBoolean("isOrdered", {locator1, locator2})
end

--[[
Retrieves the horizontal position of an element

'locator' is an element locator pointing to an element OR an element itself
--]]
function methods:getElementPositionLeft (locator)
	return self:getNumber("getElementPositionLeft", {locator})
end

--[[
Retrieves the vertical position of an element

'locator' is an element locator pointing to an element OR an element itself
--]]
function methods:getElementPositionTop (locator)
	return self:getNumber("getElementPositionTop", {locator})
end

--[[
Retrieves the width of an element

'locator' is an element locator pointing to an element
--]]
function methods:getElementWidth (locator)
	return self:getNumber("getElementWidth", {locator})
end

--[[
Retrieves the height of an element

'locator' is an element locator pointing to an element
--]]
function methods:getElementHeight (locator)
	return self:getNumber("getElementHeight", {locator})
end

--[[
Retrieves the text cursor position in the given input element or textarea; beware, this may not work perfectly on all browsers.

Specifically, if the cursor/selection has been cleared by JavaScript, this command will tend to
return the position of the last location of the cursor, even though the cursor is now gone from the page.  This is filed as SEL-243.

This method will fail if the specified element isn't an input element or textarea, or there is no cursor in the element.

'locator' is an element locator pointing to an input element or textarea
--]]
function methods:getCursorPosition (locator)
	return self:getNumber("getCursorPosition", {locator})
end

--[[
Returns the specified expression.

This is useful because of JavaScript preprocessing.
It is used to generate commands like assertExpression and waitForExpression.

'expression' is the value to return
--]]
function methods:getExpression (expression)
	return self:getString("getExpression", {expression})
end

--[[
Returns the number of nodes that match the specified xpath, eg. "//table" would give
the number of tables.

'xpath' is the xpath expression to evaluate. do NOT wrap this expression in a 'count()' function; we will do that for you.
--]]
function methods:getXpathCount (xpath)
	return self:getNumber("getXpathCount", {xpath})
end

--[[
Temporarily sets the "id" attribute of the specified element, so you can locate it in the future
using its ID rather than a slow/complicated XPath.  This ID will disappear once the page is
reloaded.

'locator' is an element locator pointing to an element
'identifier' is a string to be used as the ID of the specified element
--]]
function methods:assignId (locator, identifier)
	self:doCommand("assignId", {locator, identifier})
end

--[[
Specifies whether Selenium should use the native in-browser implementation
of XPath (if any native version is available); if you pass "false" to
this function, we will always use our pure-JavaScript xpath library.
Using the pure-JS xpath library can improve the consistency of xpath
element locators between different browser vendors, but the pure-JS
version is much slower than the native implementations.

'allow' is boolean, true means we'll prefer to use native XPath; false means we'll only use JS XPath
--]]
function methods:allowNativeXpath (allow)
	self:doCommand("allowNativeXpath", {allow})
end

--[[
Specifies whether Selenium will ignore xpath attributes that have no
value, i.e. are the empty string, when using the non-native xpath
evaluation engine. You'd want to do this for performance reasons in IE.
However, this could break certain xpaths, for example an xpath that looks
for an attribute whose value is NOT the empty string.

The hope is that such xpaths are relatively rare, but the user should
have the option of using them. Note that this only influences xpath
evaluation when using the ajaxslt engine (i.e. not "javascript-xpath").

'ignore' is boolean, true means we'll ignore attributes without valueat the expense of xpath "correctness"; false meanswe'll sacrifice speed for correctness.
--]]
function methods:ignoreAttributesWithoutValue (ignore)
	self:doCommand("ignoreAttributesWithoutValue", {ignore})
end

--[[
Runs the specified JavaScript snippet repeatedly until it evaluates to "true".
The snippet may have multiple lines, but only the result of the last line
will be considered.

Note that, by default, the snippet will be run in the runner's test window, not in the window
of your application.  To get the window of your application, you can use
the JavaScript snippet ``browserbot.getCurrentWindow()``, and then
run your JavaScript in there

'script' is the JavaScript snippet to run
'timeout' is a timeout in milliseconds, after which this command will return with an error
--]]
function methods:waitForCondition (script, timeout)
	self:doCommand("waitForCondition", {script, timeout})
end

--[[
Specifies the amount of time that Selenium will wait for actions to complete.

Actions that require waiting include "open" and the "waitFor\*" actions.

The default timeout is 30 seconds.

'timeout' is a timeout in milliseconds, after which the action will return with an error
--]]
function methods:setTimeout (timeout)
	self:doCommand("setTimeout", {timeout})
end

--[[
Waits for a new page to load.

You can use this command instead of the "AndWait" suffixes, "clickAndWait", "selectAndWait", "typeAndWait" etc.
(which are only available in the JS API).

Selenium constantly keeps track of new pages loading, and sets a "newPageLoaded"
flag when it first notices a page load.  Running any other Selenium command after
turns the flag to false.  Hence, if you want to wait for a page to load, you must
wait immediately after a Selenium command that caused a page-load.

'timeout' is a timeout in milliseconds, after which this command will return with an error
--]]
function methods:waitForPageToLoad (timeout)
	self:doCommand("waitForPageToLoad", {timeout})
end

--[[
Waits for a new frame to load.

Selenium constantly keeps track of new pages and frames loading, 
and sets a "newPageLoaded" flag when it first notices a page load.

See waitForPageToLoad for more information.

'frameAddress' is FrameAddress from the server side
'timeout' is a timeout in milliseconds, after which this command will return with an error
--]]
function methods:waitForFrameToLoad (frameAddress, timeout)
	self:doCommand("waitForFrameToLoad", {frameAddress, timeout})
end

--[[
Return all cookies of the current page under test.
--]]
function methods:getCookie ()
	return self:getString("getCookie", {})
end

--[[
Returns the value of the cookie with the specified name, or throws an error if the cookie is not present.

'name' is the name of the cookie
--]]
function methods:getCookieByName (name)
	return self:getString("getCookieByName", {name})
end

--[[
Returns true if a cookie with the specified name is present, or false otherwise.

'name' is the name of the cookie
--]]
function methods:isCookiePresent (name)
	return self:getBoolean("isCookiePresent", {name})
end

--[[
Create a new cookie whose path and domain are same with those of current page
under test, unless you specified a path for this cookie explicitly.

'nameValuePair' is name and value of the cookie in a format "name=value"
'optionsString' is options for the cookie. Currently supported options include 'path', 'max_age' and 'domain'.      the optionsString's format is "path=/path/, max_age=60, domain=.foo.com". The order of options are irrelevant, the unit      of the value of 'max_age' is second.  Note that specifying a domain that isn't a subset of the current domain will      usually fail.
--]]
function methods:createCookie (nameValuePair, optionsString)
	self:doCommand("createCookie", {nameValuePair, optionsString})
end

--[[
Delete a named cookie with specified path and domain.  Be careful; to delete a cookie, you
need to delete it using the exact same path and domain that were used to create the cookie.
If the path is wrong, or the domain is wrong, the cookie simply won't be deleted.  Also
note that specifying a domain that isn't a subset of the current domain will usually fail.

Since there's no way to discover at runtime the original path and domain of a given cookie,
we've added an option called 'recurse' to try all sub-domains of the current domain with
all paths that are a subset of the current path.  Beware; this option can be slow.  In
big-O notation, it operates in O(n\*m) time, where n is the number of dots in the domain
name and m is the number of slashes in the path.

'name' is the name of the cookie to be deleted
'optionsString' is options for the cookie. Currently supported options include 'path', 'domain'      and 'recurse.' The optionsString's format is "path=/path/, domain=.foo.com, recurse=true".      The order of options are irrelevant. Note that specifying a domain that isn't a subset of      the current domain will usually fail.
--]]
function methods:deleteCookie (name, optionsString)
	self:doCommand("deleteCookie", {name, optionsString})
end

--[[
Calls deleteCookie with recurse=true on all cookies visible to the current page.
As noted on the documentation for deleteCookie, recurse=true can be much slower
than simply deleting the cookies using a known domain/path.
--]]
function methods:deleteAllVisibleCookies ()
	self:doCommand("deleteAllVisibleCookies", {})
end

--[[
Sets the threshold for browser-side logging messages; log messages beneath this threshold will be discarded.
Valid logLevel strings are: "debug", "info", "warn", "error" or "off".
To see the browser logs, you need to
either show the log window in GUI mode, or enable browser-side logging in Selenium RC.

'logLevel' is one of the following: "debug", "info", "warn", "error" or "off"
--]]
function methods:setBrowserLogLevel (logLevel)
	self:doCommand("setBrowserLogLevel", {logLevel})
end

--[[
Creates a new "script" tag in the body of the current test window, and 
adds the specified text into the body of the command.  Scripts run in
this way can often be debugged more easily than scripts executed using
Selenium's "getEval" command.  Beware that JS exceptions thrown in these script
tags aren't managed by Selenium, so you should probably wrap your script
in try/catch blocks if there is any chance that the script will throw
an exception.

'script' is the JavaScript snippet to run
--]]
function methods:runScript (script)
	self:doCommand("runScript", {script})
end

--[[
Defines a new function for Selenium to locate elements on the page.
For example,
if you define the strategy "foo", and someone runs click("foo=blah"), we'll
run your function, passing you the string "blah", and click on the element 
that your function
returns, or throw an "Element not found" error if your function returns null.

We'll pass three arguments to your function:

*   locator: the string the user passed in
*   inWindow: the currently selected window
*   inDocument: the currently selected document


The function must return null if the element can't be found.

'strategyName' is the name of the strategy to define; this should use only   letters [a-zA-Z] with no spaces or other punctuation.
'functionDefinition' is a string defining the body of a function in JavaScript.   For example: ``return inDocument.getElementById(locator);``
--]]
function methods:addLocationStrategy (strategyName, functionDefinition)
	self:doCommand("addLocationStrategy", {strategyName, functionDefinition})
end

--[[
Saves the entire contents of the current window canvas to a PNG file.
Contrast this with the captureScreenshot command, which captures the
contents of the OS viewport (i.e. whatever is currently being displayed
on the monitor), and is implemented in the RC only. Currently this only
works in Firefox when running in chrome mode, and in IE non-HTA using
the EXPERIMENTAL "Snapsie" utility. The Firefox implementation is mostly
borrowed from the Screengrab! Firefox extension. Please see
http://www.screengrab.org and http://snapsie.sourceforge.net/ for
details.

'filename' is the path to the file to persist the screenshot as. No  filename extension will be appended by default.  Directories will not be created if they do not exist,    and an exception will be thrown, possibly by native  code.
'kwargs' is a kwargs string that modifies the way the screenshot  is captured. Example: "background=#CCFFDD" .  Currently valid options:  
*    background
    the background CSS for the HTML document. This     may be useful to set for capturing screenshots of     less-than-ideal layouts, for example where absolute     positioning causes the calculation of the canvas     dimension to fail and a black background is exposed     (possibly obscuring black text).
--]]
function methods:captureEntirePageScreenshot (filename, kwargs)
	self:doCommand("captureEntirePageScreenshot", {filename, kwargs})
end

--[[
Executes a command rollup, which is a series of commands with a unique
name, and optionally arguments that control the generation of the set of
commands. If any one of the rolled-up commands fails, the rollup is
considered to have failed. Rollups may also contain nested rollups.

'rollupName' is the name of the rollup command
'kwargs' is keyword arguments string that influences how the    rollup expands into commands
--]]
function methods:rollup (rollupName, kwargs)
	self:doCommand("rollup", {rollupName, kwargs})
end

--[[
Loads script content into a new script tag in the Selenium document. This
differs from the runScript command in that runScript adds the script tag
to the document of the AUT, not the Selenium document. The following
entities in the script content are replaced by the characters they
represent:

    &lt;
    &gt;
    &amp;

The corresponding remove command is removeScript.

'scriptContent' is the Javascript content of the script to add
'scriptTagId' is (optional) the id of the new script tag. If       specified, and an element with this id already       exists, this operation will fail.
--]]
function methods:addScript (scriptContent, scriptTagId)
	self:doCommand("addScript", {scriptContent, scriptTagId})
end

--[[
Removes a script tag from the Selenium document identified by the given
id. Does nothing if the referenced tag doesn't exist.

'scriptTagId' is the id of the script element to remove.
--]]
function methods:removeScript (scriptTagId)
	self:doCommand("removeScript", {scriptTagId})
end

--[[
Allows choice of one of the available libraries.

'libraryName' is name of the desired library Only the following three can be chosen: 
*   "ajaxslt" - Google's library
*   "javascript-xpath" - Cybozu Labs' faster library
*   "default" - The default library.  Currently the default library is "ajaxslt" .

 If libraryName isn't one of these three, then  no change will be made.
--]]
function methods:useXpathLibrary (libraryName)
	self:doCommand("useXpathLibrary", {libraryName})
end

--[[
Writes a message to the status bar and adds a note to the browser-side
log.

'context' is the message to be sent to the browser
--]]
function methods:setContext (context)
	self:doCommand("setContext", {context})
end

--[[
Sets a file input (upload) field to the file listed in fileLocator

'fieldLocator' is an element locator
'fileLocator' is a URL pointing to the specified file. Before the file  can be set in the input field (fieldLocator), Selenium RC may need to transfer the file    to the local machine before attaching the file in a web page form. This is common in selenium  grid configurations where the RC server driving the browser is not the same  machine that started the test.   Supported Browsers: Firefox ("\*chrome") only.
--]]
function methods:attachFile (fieldLocator, fileLocator)
	self:doCommand("attachFile", {fieldLocator, fileLocator})
end

--[[
Captures a PNG screenshot to the specified file.

'filename' is the absolute path to the file to be written, e.g. "c:\blah\screenshot.png"
--]]
function methods:captureScreenshot (filename)
	self:doCommand("captureScreenshot", {filename})
end

--[[
Capture a PNG screenshot.  It then returns the file as a base 64 encoded string.
--]]
function methods:captureScreenshotToString ()
	return self:getString("captureScreenshotToString", {})
end

--[[
Returns the network traffic seen by the browser, including headers, AJAX requests, status codes, and timings. When this function is called, the traffic log is cleared, so the returned content is only the traffic seen since the last call.

'type' is The type of data to return the network traffic as. Valid values are: json, xml, or plain.
--]]
function methods:captureNetworkTraffic (type)
	return self:getString("captureNetworkTraffic", {type})
end

--[[
Tells the Selenium server to add the specificed key and value as a custom outgoing request header. This only works if the browser is configured to use the built in Selenium proxy.

'key' the header name.
'value' the header value.
--]]
function methods:addCustomRequestHeader (key, value)
	return self:doCommand("addCustomRequestHeader", {key, value})
end
    
--[[
Downloads a screenshot of the browser current window canvas to a 
based 64 encoded PNG file. The \ *entire* windows canvas is captured,
including parts rendered outside of the current view port.

Currently this only works in Mozilla and when running in chrome mode.

'kwargs' is A kwargs string that modifies the way the screenshot is captured. Example: "background=#CCFFDD". This may be useful to set for capturing screenshots of less-than-ideal layouts, for example where absolute positioning causes the calculation of the canvas dimension to fail and a black background is exposed  (possibly obscuring black text).
--]]
function methods:captureEntirePageScreenshotToString (kwargs)
	return self:getString("captureEntirePageScreenshotToString", {kwargs})
end

--[[
Kills the running Selenium Server and all browser sessions.  After you run this command, you will no longer be able to send
commands to the server; you can't remotely start the server once it has been stopped.  Normally
you should prefer to run the "stop" command, which terminates the current browser session, rather than 
shutting down the entire server.
--]]
function methods:shutDownSeleniumServer ()
	self:doCommand("shutDownSeleniumServer", {})
end

--[[
Retrieve the last messages logged on a specific remote control. Useful for error reports, especially
when running multiple remote controls in a distributed environment. The maximum number of log messages
that can be retrieve is configured on remote control startup.
--]]
function methods:retrieveLastRemoteControlLogs ()
	return self:getString("retrieveLastRemoteControlLogs", {})
end

--[[
Simulates a user pressing a key (without releasing it yet) by sending a native operating system keystroke.
This function uses the java.awt.Robot class to send a keystroke; this more accurately simulates typing
a key on the keyboard.  It does not honor settings from the shiftKeyDown, controlKeyDown, altKeyDown and
metaKeyDown commands, and does not target any particular HTML element.  To send a keystroke to a particular
element, focus on the element first before running this command.

'keycode' is an integer keycode number corresponding to a java.awt.event.KeyEvent; note that Java keycodes are NOT the same thing as JavaScript keycodes!
--]]
function methods:keyDownNative (keycode)
	self:doCommand("keyDownNative", {keycode})
end

--[[
Simulates a user releasing a key by sending a native operating system keystroke.
This function uses the java.awt.Robot class to send a keystroke; this more accurately simulates typing
a key on the keyboard.  It does not honor settings from the shiftKeyDown, controlKeyDown, altKeyDown and
metaKeyDown commands, and does not target any particular HTML element.  To send a keystroke to a particular
element, focus on the element first before running this command.

'keycode' is an integer keycode number corresponding to a java.awt.event.KeyEvent; note that Java keycodes are NOT the same thing as JavaScript keycodes!
--]]
function methods:keyUpNative (keycode)
	self:doCommand("keyUpNative", {keycode})
end

--[[
Simulates a user pressing and releasing a key by sending a native operating system keystroke.
This function uses the java.awt.Robot class to send a keystroke; this more accurately simulates typing
a key on the keyboard.  It does not honor settings from the shiftKeyDown, controlKeyDown, altKeyDown and
metaKeyDown commands, and does not target any particular HTML element.  To send a keystroke to a particular
element, focus on the element first before running this command.

'keycode' is an integer keycode number corresponding to a java.awt.event.KeyEvent; note that Java keycodes are NOT the same thing as JavaScript keycodes!
--]]
function methods:keyPressNative (keycode)
	self:doCommand("keyPressNative", {keycode})
end

